<?php

namespace Drupal\config_enforce_devel\Form;

use Drupal\Core\Config\InstallStorage;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\config_enforce\ConfigEnforce;
use Drupal\config_enforce_devel\ConfigEnforceDevel;
use Drupal\config_enforce_devel\Form\ModalFormHelperTrait;
use Drupal\config_enforce_devel\Form\DevelFormHelperTrait;

/**
 * Modal form and related logic to create a new target module.
 */
class AddModuleForm extends FormBase {

  use ModalFormHelperTrait, DevelFormHelperTrait;

  const FORM_ID = 'config_enforce_devel_add_module_form';
  const MODAL_TITLE = 'Create new target module';
  const REDIRECT_ROUTE = 'config_enforce_devel.settings';

  protected $moduleName;
  protected $moduleMachineName;
  protected $moduleDescription;
  protected $modulesPath;

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $this->setSharedFormProperties($form, $form_state);

    $this->makeFormModal();
    $this->form()['actions']['send']['#value'] = $this->t('Create module');

    $this->form()['add_module'] = [
      '#type' => 'container',
      '#tree' => TRUE,
    ];
    $this->form()['add_module']['name'] = [
      '#type' => 'textfield',
      '#title' => 'Name',
      '#description' => 'The name of the new target module.',
      '#required' => TRUE,
    ];
    $this->form()['add_module']['machine_name'] = [
      '#type' => 'textfield',
      '#title' => 'Machine name',
      '#description' => 'The machine name of the new target module.',
      '#required' => TRUE,
    ];
    $this->form()['add_module']['description'] = [
      '#type' => 'textfield',
      '#title' => 'Description',
      '#description' => 'The description to include in the new target module.'
    ];
    $this->form()['add_module']['path'] = [
      '#type' => 'textfield',
      '#title' => 'Modules path',
      '#default_value' => 'modules/custom',
      '#description' => 'The path in which to create the new target module.',
      '#required' => TRUE,
    ];

    return $this->form();
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $this->setExtendedSharedFormProperties($form, $form_state);
    parent::validateForm($this->form(), $this->formState());

    // Ensure module name is unique.
    $modules = $this->getAllModules();
    $new_module = $this->formState()->getValue('add_module');
    if (in_array($new_module['name'], $modules)) {
      $this->formState()
        ->setErrorByName('add_module][name', $this->t('Module name must be unique.'));
    }
    if (array_key_exists($new_module['machine_name'], $modules)) {
      $this->formState()
        ->setErrorByName('add_module][machine_name', $this->t('Machine name must be unique.'));
    }
    // Ensure module path does not exist.
    $path = $this->getModuleDirectoryPath();
    if (is_dir($path)) {
      $this->formState()
        ->setErrorByName('add_module][path', $this->t('Module path (%path) already exists.', ['%path' => $path]));
    }
  }

  /**
   * Set form-specific properties, in addition to standard ones.
   */
  protected function setExtendedSharedFormProperties(array &$form, FormStateInterface $form_state) {
    $this->setSharedFormProperties($form, $form_state);

    $form_values = $this->formState()->getValue('add_module');
    $this->moduleName = $form_values['name'];
    $this->moduleMachineName = $form_values['machine_name'];
    $this->moduleDescription = $form_values['description'];
    $this->modulesPath = $form_values['path'];
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $this->setExtendedSharedFormProperties($form, $form_state);

    $this->createModule();
    $this->enableModule();
    $this->registerTargetModule();
    $this->initializeTargetModuleEnforcedConfigs();
  }

  /**
   * Create a new target module.
   */
  protected function createModule() {
    $this->prepareModuleDirectory();
    $this->writeInfoFile();

    $message = $this->t('%name (@machine_name) module created at %path.', [
      '%name' => $this->moduleName,
      '@machine_name' => $this->moduleMachineName,
      '%path' => $this->getModuleDirectoryPath(),
    ]);
    $this->logger('config_enforce_devel')->notice($message);
    $this->messenger()->addStatus($message);
  }

  /**
   * Ensure that the directory, in which a module info file should be written, exists.
   *
   * @param string $directory
   *   (Optional) A directory within the file path to create.
   */
  protected function prepareModuleDirectory(string $directory = '') {
    $options = FileSystemInterface::CREATE_DIRECTORY + FileSystemInterface::MODIFY_PERMISSIONS;
    $directory_path = $this->getModuleDirectoryPath();
    if (!empty($directory)) {
      $directory_path = $directory_path . '/' . $directory;
    }
    \Drupal::service('file_system')
      ->prepareDirectory($directory_path, $options);
    return $this;
  }

  /**
   * Return the path of the new target module directory.
   */
  protected function getModuleDirectoryPath() {
    return $this->modulesPath . '/' . $this->moduleMachineName;
  }

  /**
   * Write a module info file.
   */
  protected function writeInfoFile() {
    \Drupal::service('file_system')
      ->saveData($this->getInfoFileContents(), $this->getInfoFilePath());
  }

  /**
   * Create the initial enforced configs for the target module.
   */
  protected function initializeTargetModuleEnforcedConfigs() {
    $config_enforce_devel = new ConfigEnforceDevel();
    // Enforced configs settings must, themselves, be enforced.
    $config_enforce_devel
      ->setConfigName(ConfigEnforce::CONFIG_NAME . '.' . $this->moduleMachineName)
      // Enforced configs settings are specific to their target module.
      ->setTargetModule($this->moduleMachineName)
      // Enforced configs settings must be imported when the terget module is enabled.
      ->setConfigDirectory(InstallStorage::CONFIG_INSTALL_DIRECTORY)
      // Enforced configs settings must, themselves, be read-only.
      ->setEnforcementLevel(ConfigEnforce::CONFIG_ENFORCE_READONLY)
      // While read-only, the enforced configs page is the form where this'll be embedded.
      ->setConfigFormUri(Url::fromRoute('config_enforce_devel.enforced_configs')->toString())
      ->save();

    $this->writeTargetModuleEnforcedConfigs();
  }

  /**
   * Write the initial enforced configs for the target module.
   *
   * Config Devel is generally responsible for writing enforced config to disk.
   * However, here we've just initialized the target module's enforced configs
   * with settings for this very config object. Furthermore, we disable the
   * embedded forms for target module enforced configs, since these cannot be
   * changed without breaking the desired behaviour.
   * So, we have to write it ourselves, after which it should never change.
   *
   * @see initializeTargetModuleEnforcedConfigs();
   * @see DevelFormHelpoerTrait::makeEnforcedConfigsReadOnly();
   */
  protected function writeTargetModuleEnforcedConfigs() {
    // Ensure that our config directory exists.
    $this->prepareModuleDirectory(InstallStorage::CONFIG_INSTALL_DIRECTORY);

    $config_name = ConfigEnforce::CONFIG_NAME . '.' . $this->moduleMachineName;
    $config_path = $this->getModuleDirectoryPath() . '/' . InstallStorage::CONFIG_INSTALL_DIRECTORY;
    \Drupal::service('config_devel.writeback_subscriber')
      ->writeBackConfig(\Drupal::config($config_name), [$config_path]);
  }

  /**
   * Return the path of the new target module's info file.
   */
  protected function getInfoFilePath() {
    return $this->getModuleDirectoryPath() . '/' . $this->moduleMachineName . '.info.yml';
  }

  /**
   * Return a rendered the info file template.
   */
  protected function getInfoFileContents() {
    $template = drupal_get_path('module', 'config_enforce_devel') . '/templates/info.yml.twig';
    // Cast to string since twig_render_template returns a Markup object.
    return (string) twig_render_template($template, [
      'name' => $this->moduleName,
      'description' => $this->moduleDescription,
      // Needed to prevent notices when Twig debugging is enabled.
      'theme_hook_original' => 'not-applicable',
    ]);
  }

  /**
   * Enable the new target module.
   */
  protected function enableModule() {
    \Drupal::service('module_installer')->install([$this->moduleMachineName]);
  }

  /**
   * Register the new module as a target for writing config.
   */
  protected function RegisterTargetModule() {
    $config = \Drupal::configFactory()->getEditable('config_enforce_devel.settings');
    $available_modules = $config->get('available_modules');
    $available_modules[$this->moduleMachineName] = $this->moduleMachineName;
    $config->set('available_modules', $available_modules)->save();
  }

}
